function DataManager() 
{
  this.dataRoot = undefined;
  this.WPTDataCache = undefined;
}

DataManager.prototype.FillData = function( dataRoot, callback )
{
	var self = this;
	self.dataRoot = dataRoot;
	self.WPTDataCache = new WPTDataCache();
	
	$("#status-message").text("Filled data directly from .js file : " + dataRoot.values.length +" values" );
	
	callback(undefined);
};

DataManager.prototype.InitializeData = function( dataRoot, callback )
{
	var self = this;
	self.dataRoot = this.PreprocessRawData(dataRoot);
	self.WPTDataCache = new WPTDataCache();
	
	self.WPTDataCache.LoadFromStorage( function(error)
	{
		if( error )
		{
			console.error("DataManager:InitializeData : error loading WPT Data. Aborting...", error);
			callback(error);
		}
		else
		{
			self.FillWPTData( callback );
		}
	});
};

// primarily sanity checks 
DataManager.prototype.PreprocessRawData = function( rawData )
{
	console.log("DataManager:preprocessData", rawData);
	
	var testcaseTagFound = false;
	var amountOfReassignedHTTP1sRuns = 0;
	var processRecursively = function( root )
	{
		// for a description of the data model, look at UIManager.js:showDataSorted in the Oracle tool 
		// in general, it's a simple tree where each level represents a tag type (testrun, testcase, frontend, etc.) and the children names are the concrete instances of those tags that we have in the data 
		
		// so here we go breadth-first through the tree
		
		// data is a recursive datastructure
		// dataElement
		//	.tag	// name of the tag for this element (frontend, backend, protocol, etc.)
		//  .parentTagValue // name of the actual tag value for this element's parent (for frontend this could be: sitespeed_chrome, wpt_firefox, etc.)
		//	.values // the values we want to render here
		//  .children {} : associative array: {tagValue : dataElement, tagValue2 : dataElement2, etc.}
		//		= list of dataElements
		
		if( root.tag == "testcase" ) 
		{
			testcaseTagFound = true;
			
			if( Object.keys(root.children).length != 1 )
			{
				console.error("DataManager:preprocessData : for now there should only be one testcase here, namely H2priorities_H2oWithVsWithoutHttp2Prios! ", root.children);
				
				// TODO: handle better! 
				delete root.children[ Object.keys(root.children).indexOf("H2priorities_H2oParallelizationSerializationHttp2Prios") ];
			}
		}
		
		if( root.tag == "testrun" )
		{
			/*
			var wikipediaIndex = Object.keys(root.children).indexOf("H2priorities_pilot_test_realistic_wikipedia");
			if( wikipediaIndex >= 0 )
			{
				
				delete root.children[ "H2priorities_pilot_test_realistic_wikipedia" ];
				
				console.error("DataManager:preprocessData : removed wikipedia because it had only 5 runs. TODO: re-add! ", root );
				// TODO: also re-enable checking for wikipedia in PostProcessData
			} 
			*/
			
			// TODO: merge HTTP1s with others!
			/*			
			for( var childName of Object.keys(root.children) )
			{
				if( childName.indexOf("http1s") >= 0 )
					delete root.children[ childName ];
			}
			*/
		}
		
		if( root.tag == "subcase" )
		{
			var testrun = root; 
			//console.log( "Looking at testrun " + testrun.parentTagValue, root );
			if( testrun.parentTagValue.indexOf("_http1s") >= 0 )
			{
				//console.log("Found http1s testrun " + testrun.parentTagValue);
				amountOfReassignedHTTP1sRuns++;
				
				// sanity check 
				if( testrun.values.length != 1 )
				{
					if( testrun.values[0][ testrun.values[0][0].indexOf("median") ] != 0 ) // only alert if we have 2 valid values 
						console.error("DataManager:preprocessData : http1s testrun had more than 1 value!", root);
				}
				
				// sanity check 
				if( Object.keys(testrun.children).length != 1 || !testrun.children.hasOwnProperty("h2o_http1s") )
				{
					console.error("DataManager:preprocessData : http1s subcase is not called h2o_http1s or has more than 1!", testrun.children.hasOwnProperty("h2o_http1s"), root);
				}
				
				
				// all our subcases are in the same testrun, EXCEPT for the http1s subcase, which is in a testrun with the _http1s suffix
				// we want to merge that subcase with the other testrun for much easier processing 
				// we find the closest match to the _http1s subcase by removing _http1s and looking for the remaining string as prefix in another testrun (should find one!)
				// we then transfer all data from _http1s to the other one:
				// - transfer subcase to new parent .children
				// - update subcase parent and parentTagValue
				// - move .values to the new parent's .values as well (we don't use this anywhere yet, but let's be consistent shall we?)
				// - delete our parent from the list of testruns 
			
				var closestSibling = undefined;
				var siblingPrefix = testrun.parentTagValue.replace("_http1s", ""); 
				
				for( var siblingName of Object.keys(testrun.parent.children) )
				{
					if( siblingName.indexOf("_http1s") >= 0 )
						continue;
					
					if( siblingName.indexOf(siblingPrefix) == 0 )
					{
						closestSibling = testrun.parent.children[ siblingName ];
						break;
					}
				}
				
				if( !closestSibling )
				{
					console.error("DataManager:preprocessData : couldn't find a fit sibling for " + testrun.parentTagValue + " case!", testrun);
				}
				else
				{
					var newParentTestrun = closestSibling;
					var subcase = testrun.children[ "h2o_http1s" ];
					
					// - transfer subcase to new parent .children
					newParentTestrun.children[ "h2o_http1s" ] = subcase;
					
					// - update subcase parent and parentTagValue
					subcase.parent = newParentTestrun;
					subcase.parentTagValue = "h2o_http1s";//newParentTestrun.parentTagValue;
					
					// - move .values to the new parent's .values as well (we don't use this anywhere yet, but let's be consistent shall we?)
					for( var value of subcase.values )
					{
						value[ value[0].indexOf("testrun") ] = newParentTestrun.parentTagValue; // also change in the value itself, not just the global subcase
						newParentTestrun.values.push( value ); 
					}
					
					// - delete our parent from the list of testruns 
					delete testrun.parent.children[ testrun.parentTagValue ];
					
					//console.log("Re-assigned " + testrun.parentTagValue + " to " + newParentTestrun.parentTagValue, newParentTestrun, testrun.parent);
					
					/*
					// sanity check debug output 
					// we see _http1s being re-assigned multiple times, because we're quite far down the hierarchy and each path is followed independently
					// i.e. re-assignment happens for chrome and firefox, all traffic, etc. 
					var upTheChain = "";
					var current = subcase;
					while( current && current != undefined && current.length != 0 )
					{
						upTheChain += "." + current.parentTagValue;
						current = current.parent;
					}
					document.write("" + testrun.parentTagValue + "<br/>" + newParentTestrun.parentTagValue + " @ "+ upTheChain +"<br/>");
					*/
				}
			}
		}
		 
		
		for( var child of Object.keys(root.children) )
		{
			processRecursively( root.children[child] );
		}
		
		
		// TODO: merge HTTP1s runs with main testruns (take into account additional suffixes (i.e. _1, _3, etc.))
	};
	
	processRecursively( rawData );
	
	// sanity check 
	var current = rawData;
	var amountOfRemainingTestruns = -1;
	while( current && current != undefined )
	{
		if( current.tag == "testrun" )
		{
			amountOfRemainingTestruns = Object.keys(current.children).length;
			break;
		}
		
		current = current.children[ Object.keys(current.children)[0] ]; // all children should have an equal amount of testruns
	}
	console.log("DataManager:preprocessData : reassigned " + amountOfReassignedHTTP1sRuns + " HTTP1s testruns. Total testruns remaining: " + amountOfRemainingTestruns);
	
	
	if( testcaseTagFound )
	{
		alert("DataManager:preprocessData : Found root tag testcase, shouldn't happen anymore!");
	}
	
	console.log( "RAW DATA AFTER H1 MERGES ", rawData );
	
	return rawData;
}

// after WPT is fully loaded, do a final sanity check pass 
DataManager.prototype.PostProcessData = function()
{
	var self = this;
	
	// see if the data we got from influxdb is similar to that found in the WPT external link data 
	// primarily a sanity check 
	for( var value of self.dataRoot.values )
	{
		var wptData = value[ value.length - 1 ].wptData.data;
		var metricName = value[ value[0].indexOf("metric") ];
		
		if( Object.keys(wptData.runs).length != 20 ) 
		{
			if( value[ value[0].indexOf("testrun") ].indexOf("wikipedia") < 0 ) // for the wikipedia case, see PreprocessRawData
			{
				console.error("DataManager:PostProcessData : measurement has only " + Object.keys(wptData.runs).length + " runs!", value);
			}
			
			continue;
		}
		
		var WPTIndividualValues = [];
		var measurementIndices = Object.keys(wptData.runs);
		for( var measurementIndex of measurementIndices )
		{
			WPTIndividualValues.push( wptData.runs[ measurementIndex ].firstView[ metricName ] );
		}
		
		WPTIndividualValues.sort(function(a, b) {
			return a - b;
		});
		
		var influxDBMedian = value[ value[0].indexOf("median") ];
		var calculatedMedian = WPTIndividualValues[9] + ((WPTIndividualValues[10] - WPTIndividualValues[9]) / 2);
		
		if( Math.ceil(influxDBMedian) != Math.ceil(calculatedMedian) )
		{
			console.error("DataManager:PostProcessData : median mismatch between influxdb and WPT ", influxDBMedian, calculatedMedian, value);
			console.error( "" + metricName + " : " + influxDBMedian + " =? "+ calculatedMedian + " ?= " + wptData.median.firstView[metricName] +" -> " + WPTIndividualValues[9] + "__" + WPTIndividualValues[10] + " -> " + WPTIndividualValues );
		}
		
		var influxDBAverage = value[ value[0].indexOf("mean") ];
		var calculatedAverage = 0;
		for( var val of WPTIndividualValues )
		{
			if( val ) // some values are undefined for some strange reason... 
				calculatedAverage += val;
		}
		calculatedAverage /= WPTIndividualValues.length;
		
		if( Math.round(influxDBAverage) != Math.round(calculatedAverage) )
		{
			console.error("DataManager:PostProcessData : average mismatch between influxdb and WPT ", influxDBAverage, calculatedAverage, value);
			console.error( "" + metricName + " : " + influxDBAverage + " =? "+ calculatedAverage + " ?= " + wptData.average.firstView[metricName] +" ->"  + WPTIndividualValues );
		}
	}
};

DataManager.prototype.FillWPTData = function( callback )
{
	var self = this;
	var dataRoot = self.dataRoot;
	
	self.wptQueriesSent = 0;
	self.wptQueriesReceived = 0;
	
	$("#status-message").text("Filling WPT data for " + dataRoot.values.length + " values" );
	
	// we abuse the fact that ALL the values are present in the root itself as .values 
	
	for( var value of dataRoot.values ) 
	{
		(function(value){
		
		var externalIDindex = value[0].indexOf("externalID");
		if( externalIDindex < 0 )
		{
			console.error("DataManager:FetchWPTData : no externalID found for value ", value);
			return;
		}
		
		var externalID = value[externalIDindex];
		
		var wptData = self.WPTDataCache.GetLocalDataFor(externalID);
		if( !wptData )
		{
			//console.error("Couldn't find wptData for " + externalID);
			//return;
			
			//if( self.wptQueriesSent > 10 )
			//	return;
			
			self.wptQueriesSent += 1;
			
			
			self.WPTDataCache.FetchRemoteDataFor( externalID, function(result)
			{
				self.wptQueriesReceived += 1;
				 
				
				$("#status-message").text("WPT Data done for " + ((dataRoot.values.length - self.wptQueriesSent) + self.wptQueriesReceived) + "/" + dataRoot.values.length +" values" );
				
				if( result && result.data && result.statusCode == 200 )
				{
					// storing everything in the IndexedDB is way too heavy
					// we can throw away most of the result data, since we won't need them anyway 
					result = self._CleanWPTResults( result );
					
					value.push( { wptData : result } ); 
					
					// sanity check 
					if( externalID != result.data.id )
					{
						console.error("DataManager:FillWPTData : fetched data is for wrong ID! ", externalID, result );
					}
					
					self.WPTDataCache.AddToLocalData(externalID, result);
					
					console.log("WPTData received : ", value, result);
				}
				else
					console.error("DataManager:FetchWPTData : erroneous result", result, value);
				
				
				if( self.wptQueriesSent == self.wptQueriesReceived )
				{
					$("#status-message").text("WPT Data done for all " + dataRoot.values.length +" values" );
					self.WPTDataCache.PersistLocalData();
					
					// we're all done, we can call the callback!
					callback(undefined);
				}
			});
		}
		else
		{
			// sanity check 
			if( externalID != wptData.data.id )
			{
				console.error("DataManager:FillWPTData : filled data is for wrong ID! ", externalID, result );
			}
					
			value.push( { wptData : wptData } );
		}
		
		})(value);// wrap in function so we can bring value into the inner scope (by default, it stays in the global for-loop scope and is overridden)
	}
	
	// can happen if we fill everything from local db
	if( self.wptQueriesSent == self.wptQueriesReceived )
	{
		$("#status-message").text("WPT Data done for all " + dataRoot.values.length +" values" );
		self.PostProcessData();
		callback(undefined);
	}
}

DataManager.prototype._CleanWPTResults = function( results )
{
	// only keep the values we really care about. The rest is discarded so we don't have to store huge amounts of data in the browser. 
	var keepKeys = [];
	keepKeys.push("SpeedIndex");
	keepKeys.push("loadEventEnd");
	keepKeys.push("TTFB");
	keepKeys.push("domContentLoadedEventEnd");
	keepKeys.push("priorityStreams");
	keepKeys.push("firstPaint");
	keepKeys.push("lastVisualChange");
	keepKeys.push("render");
	keepKeys.push("visualComplete");
	keepKeys.push("fullyLoaded");
	
	var runs = results.data.runs;
	for( var runName of Object.keys(runs) )
	{
		var run = runs[runName].firstView;
		
		var runKeys = Object.keys(run);
		for( var runKey of runKeys )
		{
			if( keepKeys.indexOf(runKey) < 0 )
				delete run[ runKey ];
		}
	}
	
	delete results.data.standardDeviation;
	
	return results;
};


/*
DataManager.prototype.FetchWPTData2 = function( callback )
{
	var self = this;
	var dataRoot = self.dataRoot;
	
	var currentValueIndex = -1;
	
	var getCallback = function(result)
	{
		currentValueIndex++;
		
		if( currentValueIndex < dataRoot.values.length )
		{
			var value = dataRoot.values[currentValueIndex];
			
			var externalIDindex = value[0].indexOf("externalID");
			if( externalIDindex < 0 )
			{
				console.error("DataManager:FetchWPTData : no externalID found for value ", value);
				getCallback(currentValueIndex);
				return;
			}
			
			var url = "http://pro-flow-win-1.edm.uhasselt.be/results.php?test=";
			url += value[externalIDindex];
			url += "&f=json&requests=0";
		
			$("#status-message").text("Fetching WPT data for " + currentValueIndex + "/" + dataRoot.values.length +" values" );
			$.get(url, getCallback);
		}
		
	};
	
	getCallback("first");
	
};
*/